# ------------------------------------------------------------------------------
#  es7s/commons
#  (c) 2023 A. Shavykin <0.delameter@gmail.com>
# ------------------------------------------------------------------------------
import pytermor as pt


# https://wakatime.com/colors/languages

# a={}; k=""; for (var col of jQuery("span.editor-icon")){ for (s of [...$("div",col)]) {
# d=s.innerText; if (k) { a[d]=k; k=null; } else { k = d; } } }; console.log(a);


class PLangColor(pt.RealColor, pt.RenderColor, pt.ResolvableColor["PLangColor"]):
    # fmt: off
    MAP = {
        "1C Enterprise":                  0x814ccc,
        "2-Dimensional Array":            0x38761d,
        "4D":                             0x004289,
        "ABAP CDS":                       0x555e25,
        "ABAP":                           0xe8274b,
        "AGS Script":                     0xb9d9ff,
        "AIDL":                           0x34eb6b,
        "AL":                             0x3aa2b5,
        "AMPL":                           0xe6efbb,
        "ANTLR":                          0x9dc3ff,
        "API Blueprint":                  0x2acca8,
        "APL":                            0x5a8164,
        "ASP.NET":                        0x9400ff,
        "ATS":                            0x1ac620,
        "ActionScript":                   0x882b0f,
        "Ada":                            0x02f88c,
        "Adblock Filter List":            0x800000,
        "Adobe Font Metrics":             0xfa0f00,
        "Agda":                           0x315665,
        "Alloy":                          0x64c800,
        "Alpine Abuild":                  0x0d597f,
        "Altium Designer":                0xa89663,
        "AngelScript":                    0xc7d7dc,
        "Ant Build System":               0xa9157e,
        "Antlers":                        0xff269e,
        "ApacheConf":                     0xd12127,
        "Apex":                           0x1797c0,
        "Apollo Guidance Computer":       0x0b3d91,
        "AppleScript":                    0x101f1f,
        "Arc":                            0xaa2afe,
        "AsciiDoc":                       0x73a0c5,
        "AspectJ":                        0xa957b0,
        "Assembly":                       0x6e4c13,
        "Astro":                          0xff5a03,
        "Asymptote":                      0xff0000,
        "Augeas":                         0x9cc134,
        "AutoHotkey":                     0x6594b9,
        "AutoIt":                         0x1c3552,
        "Avro IDL":                       0x0040ff,
        "Awk":                            0xc30e9b,
        "BASIC":                          0xff0000,
        "Ballerina":                      0xff5000,
        "Batchfile":                      0xc1f12e,
        "Beef":                           0xa52f4e,
        "Berry":                          0x15a13c,
        "BibTeX":                         0x778899,
        "Bicep":                          0x519aba,
        "Bikeshed":                       0x5562ac,
        "Bison":                          0x6a463f,
        "BitBake":                        0x00bce4,
        "Blade":                          0xf7523f,
        "BlitzBasic":                     0x00ffae,
        "BlitzMax":                       0xcd6400,
        "Bluespec":                       0x12223c,
        "Boo":                            0xd4bec1,
        "Boogie":                         0xc80fa0,
        "Brainfuck":                      0x2f2530,
        "BrighterScript":                 0x66aabb,
        "Brightscript":                   0x662d91,
        "Browserslist":                   0xffd539,
        "C":                              0x555555,
        "C#":                             0x178600,
        "C++":                            0xf34b7d,
        "CAP CDS":                        0x0092d1,
        "CLIPS":                          0x00a300,
        "CMake":                          0xda3434,
        "COLLADA":                        0xf1a42b,
        "CSON":                           0x244776,
        "CSS":                            0x563d7c,
        "CSV":                            0x237346,
        "CUE":                            0x5886e1,
        "CWeb":                           0x00007a,
        "Cabal Config":                   0x483465,
        "Cadence":                        0x00ef8b,
        "Cairo":                          0xff4a48,
        "CameLIGO":                       0x3be133,
        "Cap'n Proto":                    0xc42727,
        "Ceylon":                         0xdfa535,
        "Chapel":                         0x8dc63f,
        "ChucK":                          0x3f8000,
        "Circom":                         0x707575,
        "Cirru":                          0xccccff,
        "Clarion":                        0xdb901e,
        "Clarity":                        0x5546ff,
        "Classic ASP":                    0x6a40fd,
        "Clean":                          0x3f85af,
        "Click":                          0xe4e6f3,
        "Clojure":                        0xdb5855,
        "Closure Templates":              0x0d948f,
        "Cloud Firestore Security Rules": 0xffa000,
        "CodeQL":                         0x140f46,
        "CoffeeScript":                   0x244776,
        "ColdFusion CFC":                 0xed2cd6,
        "ColdFusion":                     0xed2cd6,
        "Common Lisp":                    0x3fb68b,
        "Common Workflow Language":       0xb5314c,
        "Component Pascal":               0xb0ce4e,
        "Coq":                            0xd0b68c,
        "Crystal":                        0x000100,
        "Csound Document":                0x1a1a1a,
        "Csound Score":                   0x1a1a1a,
        "Csound":                         0x1a1a1a,
        "Cuda":                           0x3a4e3a,
        "Curry":                          0x531242,
        "Cypher":                         0x34c0eb,
        "Cython":                         0xfedf5b,
        "D":                              0xba595e,
        "DM":                             0x447265,
        "Dafny":                          0xffec25,
        "Darcs Patch":                    0x8eff23,
        "Dart":                           0x00b4ab,
        "DataWeave":                      0x003a52,
        "Debian Package Control File":    0xd70751,
        "DenizenScript":                  0xfbee96,
        "Dhall":                          0xdfafff,
        "DirectX 3D File":                0xaace60,
        "Dockerfile":                     0x384d54,
        "Dogescript":                     0xcca760,
        "Dotenv":                         0xe5d559,
        "Dylan":                          0x6c616e,
        "E":                              0xccce35,
        "ECL":                            0x8a1267,
        "ECLiPSe":                        0x001d9d,
        "EJS":                            0xa91e50,
        "EQ":                             0xa78649,
        "Earthly":                        0x2af0ff,
        "Easybuild":                      0x069406,
        "Ecere Projects":                 0x913960,
        "Ecmarkup":                       0xeb8131,
        "EditorConfig":                   0xfff1f2,
        "Eiffel":                         0x4d6977,
        "Elixir":                         0x6e4a7e,
        "Elm":                            0x60b5cc,
        "Elvish Transcript":              0x55bb55,
        "Elvish":                         0x55bb55,
        "Emacs Lisp":                     0xc065db,
        "EmberScript":                    0xfff4f3,
        "Erlang":                         0xb83998,
        "Euphoria":                       0xff790b,
        "F#":                             0xb845fc,
        "F*":                             0x572e30,
        "FIGlet Font":                    0xffddbb,
        "FLUX":                           0x88ccff,
        "Factor":                         0x636746,
        "Fancy":                          0x7b9db4,
        "Fantom":                         0x14253c,
        "Faust":                          0xc37240,
        "Fennel":                         0xfff3d7,
        "Filebench WML":                  0xf6b900,
        "Fluent":                         0xffcc33,
        "Forth":                          0x341708,
        "Fortran Free Form":              0x4d41b1,
        "Fortran":                        0x4d41b1,
        "FreeBasic":                      0x141ac9,
        "FreeMarker":                     0x0050b2,
        "Frege":                          0x00cafe,
        "Futhark":                        0x5f021f,
        "G-code":                         0xd08cf2,
        "GAML":                           0xffc766,
        "GAMS":                           0xf49a22,
        "GAP":                            0x0000cc,
        "GCC Machine Description":        0xffcfab,
        "GDScript":                       0x355570,
        "GEDCOM":                         0x003058,
        "GLSL":                           0x5686a5,
        "GSC":                            0xff6800,
        "Game Maker Language":            0x71b417,
        "Gemfile.lock":                   0x701516,
        "Gemini":                         0xff6900,
        "Genero Forms":                   0xd8df39,
        "Genero":                         0x63408e,
        "Genie":                          0xfb855d,
        "Genshi":                         0x951531,
        "Gentoo Ebuild":                  0x9400ff,
        "Gentoo Eclass":                  0x9400ff,
        "Gerber Image":                   0xd20b00,
        "Gherkin":                        0x5b2063,
        "Git Attributes":                 0xf44d27,
        "Git Config":                     0xf44d27,
        "Git Revision List":              0xf44d27,
        "Gleam":                          0xffaff3,
        "Glyph":                          0xc1ac7f,
        "Gnuplot":                        0xf0a9f0,
        "Go Checksums":                   0x00add8,
        "Go Module":                      0x00add8,
        "Go":                             0x00add8,
        "Godot Resource":                 0x355570,
        "Golo":                           0x88562a,
        "Gosu":                           0x82937f,
        "Grace":                          0x615f8b,
        "Gradle":                         0x02303a,
        "Grammatical Framework":          0xff0000,
        "GraphQL":                        0xe10098,
        "Graphviz (DOT)":                 0x2596be,
        "Groovy Server Pages":            0x4298b8,
        "Groovy":                         0x4298b8,
        "HAProxy":                        0x106da9,
        "HCL":                            0x844fba,
        "HLSL":                           0xaace60,
        "HOCON":                          0x9ff8ee,
        "HTML":                           0xe34c26,
        "HTML+ECR":                       0x2e1052,
        "HTML+EEX":                       0x6e4a7e,
        "HTML+ERB":                       0x701516,
        "HTML+PHP":                       0x4f5d95,
        "HTML+Razor":                     0x512be4,
        "HTTP":                           0x005c9c,
        "HXML":                           0xf68712,
        "Hack":                           0x878787,
        "Haml":                           0xece2a9,
        "Handlebars":                     0xf7931e,
        "Harbour":                        0x0e60e3,
        "Haskell":                        0x5e5086,
        "Haxe":                           0xdf7900,
        "HiveQL":                         0xdce200,
        "HolyC":                          0xffefaf,
        "Hy":                             0x7790b2,
        "IDL":                            0xa3522f,
        "IGOR Pro":                       0x0000cc,
        "INI":                            0xd1dbe0,
        "Idris":                          0xb30000,
        "Ignore List":                    0x000000,
        "ImageJ Macro":                   0x99aaff,
        "Imba":                           0x16cec6,
        "Inno Setup":                     0x264b99,
        "Io":                             0xa9188d,
        "Ioke":                           0x078193,
        "Isabelle ROOT":                  0xfefe00,
        "Isabelle":                       0xfefe00,
        "J":                              0x9eedff,
        "JAR Manifest":                   0xb07219,
        "JCL":                            0xd90e09,
        "JFlex":                          0xdbca00,
        "JSON with Comments":             0x292929,
        "JSON":                           0x292929,
        "JSON5":                          0x267cb9,
        "JSONLD":                         0x0c479c,
        "JSONiq":                         0x40d47e,
        "Janet":                          0x0886a5,
        "Jasmin":                         0xd03600,
        "Java Properties":                0x2a6277,
        "Java Server Pages":              0x2a6277,
        "Java":                           0xb07219,
        "JavaScript":                     0xf1e05a,
        "JavaScript+ERB":                 0xf1e05a,
        "Jest Snapshot":                  0x15c213,
        "JetBrains MPS":                  0x21d789,
        "Jinja":                          0xa52a22,
        "Jison Lex":                      0x56b3cb,
        "Jison":                          0x56b3cb,
        "Jolie":                          0x843179,
        "Jsonnet":                        0x0064bd,
        "Julia":                          0xa270ba,
        "Jupyter Notebook":               0xda5b0b,
        "Just":                           0x384d54,
        "KRL":                            0x28430a,
        "Kaitai Struct":                  0x773b37,
        "KakouneScript":                  0x6f8042,
        "KerboScript":                    0x41adf0,
        "KiCad Layout":                   0x2f4aab,
        "KiCad Legacy Layout":            0x2f4aab,
        "KiCad Schematic":                0x2f4aab,
        "Kotlin":                         0xa97bff,
        "LFE":                            0x4c3023,
        "LLVM":                           0x185619,
        "LOLCODE":                        0xcc9900,
        "LSL":                            0x3d9970,
        "LabVIEW":                        0xfede06,
        "Lark":                           0x2980b9,
        "Lasso":                          0x999999,
        "Latte":                          0xf2a542,
        "Less":                           0x1d365d,
        "Lex":                            0xdbca00,
        "LigoLANG":                       0x0e74ff,
        "LilyPond":                       0x9ccc7c,
        "Liquid":                         0x67b8de,
        "Literate Agda":                  0x315665,
        "Literate CoffeeScript":          0x244776,
        "Literate Haskell":               0x5e5086,
        "LiveScript":                     0x499886,
        "Logtalk":                        0x295b9a,
        "LookML":                         0x652b81,
        "Lua":                            0x000080,
        "MATLAB":                         0xe16737,
        "MAXScript":                      0x00a6a6,
        "MLIR":                           0x5ec8db,
        "MQL4":                           0x62a8d6,
        "MQL5":                           0x4a76b8,
        "MTML":                           0xb7e1f4,
        "Macaulay2":                      0xd8ffff,
        "Makefile":                       0x427819,
        "Mako":                           0x7e858d,
        "Markdown":                       0x083fa1,
        "Marko":                          0x42bff2,
        "Mask":                           0xf97732,
        "Mathematica":                    0xdd1100,
        "Max":                            0xc4a79c,
        "Mercury":                        0xff2b2b,
        "Mermaid":                        0xff3670,
        "Meson":                          0x007800,
        "Metal":                          0x8f14e9,
        "MiniYAML":                       0xff1111,
        "Mint":                           0x02b046,
        "Mirah":                          0xc7a938,
        "Modelica":                       0xde1d31,
        "Modula-2":                       0x10253f,
        "Modula-3":                       0x223388,
        "Monkey C":                       0x8d6747,
        "MoonScript":                     0xff4585,
        "Motoko":                         0xfbb03b,
        "Motorola 68K Assembly":          0x005daa,
        "Move":                           0x4a137a,
        "Mustache":                       0x724b3b,
        "NCL":                            0x28431f,
        "NPM Config":                     0xcb3837,
        "NWScript":                       0x111522,
        "Nasal":                          0x1d2c4e,
        "Nearley":                        0x990000,
        "Nemerle":                        0x3d3c6e,
        "NetLinx":                        0x0aa0ff,
        "NetLinx+ERB":                    0x747faa,
        "NetLogo":                        0xff6375,
        "NewLisp":                        0x87aed7,
        "Nextflow":                       0x3ac486,
        "Nginx":                          0x009639,
        "Nim":                            0xffc200,
        "Nit":                            0x009917,
        "Nix":                            0x7e7eff,
        "Nu":                             0xc9df40,
        "NumPy":                          0x9c8af9,
        "Nunjucks":                       0x3d8137,
        "OASv2-json":                     0x85ea2d,
        "OASv2-yaml":                     0x85ea2d,
        "OASv3-json":                     0x85ea2d,
        "OASv3-yaml":                     0x85ea2d,
        "OCaml":                          0x3be133,
        "ObjectScript":                   0x424893,
        "Objective-C":                    0x438eff,
        "Objective-C++":                  0x6866fb,
        "Objective-J":                    0xff0c5a,
        "Odin":                           0x60affe,
        "Omgrofl":                        0xcabbff,
        "Opal":                           0xf7ede0,
        "Open Policy Agent":              0x7d9199,
        "OpenAPI Specification v2":       0x85ea2d,
        "OpenAPI Specification v3":       0x85ea2d,
        "OpenCL":                         0xed2e2d,
        "OpenEdge ABL":                   0x5ce600,
        "OpenQASM":                       0xaa70ff,
        "OpenSCAD":                       0xe5cd45,
        "Option List":                    0x476732,
        "Org":                            0x77aa99,
        "Oxygene":                        0xcdd0e3,
        "Oz":                             0xfab738,
        "P4":                             0x7055b5,
        "PDDL":                           0x0d00ff,
        "PEG.js":                         0x234d6b,
        "PHP":                            0x4f5d95,
        "PLSQL":                          0xdad8d8,
        "PLpgSQL":                        0x336790,
        "POV-Ray SDL":                    0x6bac65,
        "Pan":                            0xcc0000,
        "Papyrus":                        0x6600cc,
        "Parrot":                         0xf3ca0a,
        "Pascal":                         0xe3f171,
        "Pawn":                           0xdbb284,
        "Pep8":                           0xc76f5b,
        "Perl":                           0x0298c3,
        "PicoLisp":                       0x6067af,
        "PigLatin":                       0xfcd7de,
        "Pike":                           0x005390,
        "PlantUML":                       0xfbbd16,
        "PogoScript":                     0xd80074,
        "Polar":                          0xae81ff,
        "Portugol":                       0xf8bd00,
        "PostCSS":                        0xdc3a0c,
        "PostScript":                     0xda291c,
        "PowerBuilder":                   0x8f0f8d,
        "PowerShell":                     0x012456,
        "Prisma":                         0x0c344b,
        "Processing":                     0x0096d8,
        "Procfile":                       0x3b2f63,
        "Prolog":                         0x74283c,
        "Promela":                        0xde0000,
        "Propeller Spin":                 0x7fa2a7,
        "Pug":                            0xa86454,
        "Puppet":                         0x302b6d,
        "PureBasic":                      0x5a6986,
        "PureScript":                     0x1d222d,
        "Pyret":                          0xee1e10,
        "Python console":                 0x3572a5,
        "Python traceback":               0x3572a5,
        "Python":                         0x3572a5,
        "Q#":                             0xfed659,
        "QML":                            0x44a51c,
        "Qt Script":                      0x00b841,
        "Quake":                          0x882233,
        "R":                              0x198ce7,
        "RAML":                           0x77d9fb,
        "RDoc":                           0x701516,
        "REXX":                           0xd90e09,
        "RMarkdown":                      0x198ce7,
        "RPGLE":                          0x2bde21,
        "RUNOFF":                         0x665a4e,
        "Racket":                         0x3c5caa,
        "Ragel":                          0x9d5200,
        "Raku":                           0x0000fb,
        "Rascal":                         0xfffaa0,
        "ReScript":                       0xed5051,
        "Reason":                         0xff5847,
        "ReasonLIGO":                     0xff5847,
        "Rebol":                          0x358a5b,
        "Record Jar":                     0x0673ba,
        "Red":                            0xf50000,
        "Regular Expression":             0x009a00,
        "Ren'Py":                         0xff7f7f,
        "Ring":                           0x2d54cb,
        "Riot":                           0xa71e49,
        "RobotFramework":                 0x00c0b5,
        "Roff Manpage":                   0xecdebe,
        "Roff":                           0xecdebe,
        "Rouge":                          0xcc0088,
        "RouterOS Script":                0xde3941,
        "Ruby":                           0x701516,
        "Rust":                           0xdea584,
        "SAS":                            0xb34936,
        "SCSS":                           0xc6538c,
        "SPARQL":                         0x0c4597,
        "SQF":                            0x3f3f3f,
        "SQL":                            0xe38c00,
        "SQLPL":                          0xe38c00,
        "SRecode Template":               0x348a34,
        "STL":                            0x373b5e,
        "SVG":                            0xff9900,
        "SaltStack":                      0x646464,
        "Sass":                           0xa53b70,
        "Scala":                          0xc22d40,
        "Scaml":                          0xbd181a,
        "Scenic":                         0xfdc700,
        "Scheme":                         0x1e4aec,
        "Scilab":                         0xca0f21,
        "Self":                           0x0579aa,
        "ShaderLab":                      0x222c37,
        "Shell":                          0x89e051,
        "ShellCheck Config":              0xcecfcb,
        "Shen":                           0x120f14,
        "Simple File Verification":       0xc9bfed,
        "Singularity":                    0x64e6ad,
        "Slash":                          0x007eff,
        "Slice":                          0x003fa2,
        "Slim":                           0x2b2b2b,
        "SmPL":                           0xc94949,
        "Smalltalk":                      0x596706,
        "Smarty":                         0xf0c040,
        "Smithy":                         0xc44536,
        "Snakemake":                      0x419179,
        "Solidity":                       0xaa6746,
        "SourcePawn":                     0xf69e1d,
        "Squirrel":                       0x800000,
        "Stan":                           0xb2011d,
        "Standard ML":                    0xdc566d,
        "Starlark":                       0x76d275,
        "Stata":                          0x1a5f91,
        "StringTemplate":                 0x3fb34f,
        "Stylus":                         0xff6347,
        "SubRip Text":                    0x9e0101,
        "SugarSS":                        0x2fcc9f,
        "SuperCollider":                  0x46390b,
        "Svelte":                         0xff3e00,
        "Sway":                           0xdea584,
        "Swift":                          0xf05138,
        "SystemVerilog":                  0xdae1c2,
        "TI Program":                     0xa0aa87,
        "TLA":                            0x4b0079,
        "TOML":                           0x9c4221,
        "TSQL":                           0xe38c00,
        "TSV":                            0x237346,
        "TSX":                            0x3178c6,
        "TXL":                            0x0178b8,
        "Talon":                          0x333333,
        "Tcl":                            0xe4cc98,
        "TeX":                            0x3d6117,
        "Terra":                          0x00004c,
        "TextMate Properties":            0xdf66e4,
        "Textile":                        0xffe7ac,
        "Thrift":                         0xd12127,
        "Turing":                         0xcf142b,
        "Twig":                           0xc1d026,
        "TypeScript":                     0x3178c6,
        "Unified Parallel C":             0x4e3617,
        "Unity3D Asset":                  0x222c37,
        "Unknown Language":               0x1f9aef,
        "Uno":                            0x9933cc,
        "UnrealScript":                   0xa54c4d,
        "UrWeb":                          0xccccee,
        "V":                              0x4f87c4,
        "VBA":                            0x867db1,
        "VBScript":                       0x15dcdc,
        "VCL":                            0x148aa8,
        "VHDL":                           0xadb2cb,
        "Vala":                           0xa56de2,
        "Valve Data Format":              0xf26025,
        "Velocity Template Language":     0x507cff,
        "Verilog":                        0xb2b7f8,
        "Vim Help File":                  0x199f4b,
        "Vim Script":                     0x199f4b,
        "Vim Snippet":                    0x199f4b,
        "Visual Basic .NET":              0x945db7,
        "Visual Basic 6.0":               0x2c6353,
        "Volt":                           0x1f1f1f,
        "Vue":                            0x41b883,
        "Vyper":                          0x2980b9,
        "Web Ontology Language":          0x5b70bd,
        "WebAssembly":                    0x04133b,
        "Whiley":                         0xd5c397,
        "Wikitext":                       0xfc5757,
        "Windows Registry Entries":       0x52d5ff,
        "Witcher Script":                 0xff0000,
        "Wollok":                         0xa23738,
        "World of Warcraft Addon Data":   0xf7e43f,
        "Wren":                           0x383838,
        "X10":                            0x4b6bef,
        "XC":                             0x99da07,
        "XML Property List":              0x0060ac,
        "XML":                            0x0060ac,
        "XQuery":                         0x5232e7,
        "XSLT":                           0xeb8ceb,
        "Xojo":                           0x81bd41,
        "Xonsh":                          0x285eef,
        "Xtend":                          0x24255d,
        "YAML":                           0xcb171e,
        "YARA":                           0x220000,
        "YASnippet":                      0x32ab90,
        "Yacc":                           0x4b6c4b,
        "Yul":                            0x794932,
        "ZAP":                            0x0d665e,
        "ZIL":                            0xdc75e5,
        "ZenScript":                      0x00bcd1,
        "Zephir":                         0x118f9e,
        "Zig":                            0xec915c,
        "Zimpl":                          0xd67711,
        "eC":                             0x913960,
        "fish":                           0x4aae47,
        "hoon":                           0x00b171,
        "jq":                             0xc7254e,
        "kvlang":                         0x1da6e0,
        "mIRC Script":                    0x3d57c3,
        "mcfunction":                     0xe22837,
        "mupad":                          0x244963,
        "nanorc":                         0x2d004d,
        "nesC":                           0x94b0c7,
        "ooc":                            0xb0b77e,
        "q":                              0x0040cd,
        "reStructuredText":               0x141414,
        "sed":                            0x64b970,
        "wdl":                            0x42f1f4,
        "wisp":                           0x7582d1,
        "xBase":                          0x403a40,
        
        "diff":                           0xf0c0c0,
        "text":                           0xc0c0c0,
    }

    NAME_NORMALIZATION_MAP = {
        '*': 'star',
        '#': 'sharp',
        '+': 'plus',
    }
    # fmt: on

    def __init__(self, value: pt.IColorValue | int, name: str = None):
        pt.RealColor.__init__(self, value)
        pt.ResolvableColor.__init__(self, name, approx=True, register=True)

    @classmethod
    def get_longest_name(cls) -> int:
        return max(map(len, cls.MAP.keys()))

    to_sgr = pt.ColorRGB.to_sgr
    to_tmux = pt.ColorRGB.to_tmux

    def repr_attrs(self, verbose: bool = True) -> str:
        return f'{self.format_value("#")}({self.name})'

    @classmethod
    def _load(cls):
        for name, value in cls.MAP.items():
            PLangColor(value, name.lower())



# ------------------------------------------------------------------------------
#  es7s/commons
#  (c) 2021-2023 A. Shavykin <0.delameter@gmail.com>
# ------------------------------------------------------------------------------

from ._version import __updated__ as PKG_UPDATED
from ._version import __version__ as PKG_VERSION

from .common import autogen as autogen
from .common import bcs as bcs
from .common import FinalSingleton as FinalSingleton
from .common import lcm as lcm
from .common import median as median
from .common import now as now
from .common import nowf as nowf
from .common import percentile as percentile
from .common import Regex as Regex
from .gradient import deque_ext as deque_ext
from .gradient import GimpGradientReader as GimpGradientReader
from .gradient import Gradient as Gradient
from .gradient import GradientPoint as GradientPoint
from .gradient import GradientSegment as GradientSegment
from .gradient import IGradientReader as IGradientReader
from .plang import PLangColor as PLangColor
from .prof import measure as measure
from .progressbar import DummyProgressBar as DummyProgressBar
from .progressbar import ProgressBar as ProgressBar
from .pt_ import AdaptiveFragment as AdaptiveFragment
from .pt_ import CompositeCompressor as CompositeCompressor
from .pt_ import DisposableComposite as DisposableComposite
from .pt_ import filtere as filtere
from .pt_ import filterev as filterev
from .pt_ import format_attrs as format_attrs
from .pt_ import format_path as format_path
from .pt_ import isempty as isempty
from .pt_ import joincoal as joinScoal
from .scale import FULL_BLOCK as FULL_BLOCK
from .scale import get_partial_hblock as get_partial_hblock
from .scale import Scale as Scale
from .separator import FILE_SEPARATOR as FILE_SEPARATOR
from .separator import get_separator as get_separator
from .separator import GROUP_SEPARATOR as GROUP_SEPARATOR
from .separator import RECORD_SEPARATOR as RECORD_SEPARATOR
from .separator import SEPARATORS as SEPARATORS
from .separator import UNIT_SEPARATOR as UNIT_SEPARATOR
from .spinner import Spinner as Spinner
from .spinner import SpinnerBrailleSquare as SpinnerBrailleSquare
from .spinner import SpinnerBrailleSquareCenter as SpinnerBrailleSquareCenter
from .spinner import SpinnerBrailleSquareFill as SpinnerBrailleSquareFill
from .structx import DoublyLinkedNode as DoublyLinkedNode
from .structx import RingList as RingList
from .strutil import NamedGroupsRefilter as NamedGroupsRefilter
from .strutil import re_unescape as re_unescape
from .strutil import RegexValRefilter as RegexValRefilter
from .strutil import SUBSCRIPT_TRANS as SUBSCRIPT_TRANS
from .strutil import SUPERSCRIPT_TRANS as SUPERSCRIPT_TRANS
from .strutil import to_subscript as to_subscript
from .strutil import to_superscript as to_superscript
from .strutil import Transmap as Transmap
from .termstate import InputMode as InputMode
from .termstate import TerminalState as TerminalState
from .totalsize import total_size as total_size
from .weather import DynamicIcon as DynamicIcon
from .weather import get_wicon
from .weather import justify_wicon
from .weather import WEATHER_ICON_SETS as WEATHER_ICON_SETS
from .weather import WEATHER_ICON_TERMINATOR as WEATHER_ICON_TERMINATOR
from .weather import WEATHER_SYMBOL_PLAIN as WEATHER_SYMBOL_PLAIN
from .weather import WeatherIconSet as WeatherIconSet
from .weather import WIND_DIRECTION as WIND_DIRECTION
from .weather import WWO_CODE as WWO_CODE



# ------------------------------------------------------------------------------
#  es7s/commons
#  (c) 2023 A. Shavykin <0.delameter@gmail.com>
# ------------------------------------------------------------------------------
import math
import sys
import time
import typing
import typing as t
from io import StringIO

import pytermor as pt
from .scale import FULL_BLOCK, get_partial_hblock


class DummyProgressBar:
    def __init__(self, *_, **__):
        self._label = ""
        self._count = 0

    def init_tasks(self, tasks_amount: int = None, task_num: int = 1):
        ...

    def next_task(self, task_label: str = None):
        self._label = task_label
        self._count += 1

    def init_steps(self, steps_amount: int = None, step_num: int = 0):
        ...

    def next_step(self, step_label: str = None):
        self._label = step_label
        self._count += 1

    def render(self):
        sys.stderr.write(f"{pt.get_qname(self)} [{self._count}] {self._label}\n")

    def close(self, *args, **kwargs):
        ...


class ProgressBar:
    BAR_WIDTH = 5
    LABEL_PAD = 2

    MAX_FRAME_RATE = 16
    FRAME_INTERVAL_SEC = 1 / MAX_FRAME_RATE
    TERM_WIDTH_QUERY_INTERVAL_SEC = 5
    PERSIST_MIN_INTERVAL_SEC = 5

    SGR_RESET = pt.SeqIndex.RESET.assemble()
    CSI_CHA1 = pt.make_set_cursor_column(1).assemble()
    CSI_EL0 = pt.make_clear_line_after_cursor().assemble()
    OUT_START = CSI_CHA1 + CSI_EL0
    OUT_END = pt.SeqIndex.BG_COLOR_OFF.assemble()

    FIELD_SEP = " "
    ICON = "◆"
    NUM_DELIM = "/"
    BORDER_LEFT_CHAR = "▕"
    BORDER_RIGHT_CHAR = "▏"

    def __init__(
        self,
        renderer: pt.IRenderer,
        io: t.IO,
        theme_color: pt.Color,
        tasks_amount=1,
        task_num=1,
        task_label="Working",
        steps_amount=0,
        step_num=0,
        step_label="...",
        print_step_num=True,
    ):
        self._last_persist_ts: int | None = None
        self._created_at = time.monotonic_ns()

        self._renderer = renderer
        self._io = io
        self._output_buffer = _OutputBuffer()
        self._styles = _PBarStyles(theme_color)

        self._tasks_amount: int = tasks_amount
        self._task_num: int = task_num
        self._task_label: str = task_label

        self._steps_amount = steps_amount
        self._step_num = step_num
        self._step_label: str = step_label

        self._print_step_num = print_step_num

        self._next_frame_ts: int | None = None
        self._icon_frame = 0

        self._max_label_len: int | None = None
        self._last_term_width_query_ts: int | None = None

    @property
    def is_format_allowed(self) -> bool:
        return self._renderer.is_format_allowed

    def init_tasks(self, tasks_amount: int = None, task_num: int = 1):
        if tasks_amount is not None:
            self._tasks_amount = tasks_amount
        if task_num is not None:
            self._task_num = task_num
            self._steps_amount = 0
            self._step_num = 0
        self._task_num = min(self._task_num, self._get_max_task_num())

    def next_task(self, task_label: str = None, render=True):
        if task_label is not None:
            self._task_label = task_label
        self.init_tasks(task_num=self._task_num + 1)
        if render:
            self.render()

    def init_steps(self, steps_amount: int = None, step_num: int = 0):
        if steps_amount is not None:
            self._steps_amount = steps_amount
        if step_num is not None:
            self._step_num = step_num
        self._step_num = min(self._step_num, self._get_max_step_num())

    def next_step(self, step_label: str = None, render=True):
        if step_label is not None:
            self._step_label = step_label
        self.init_steps(step_num=self._step_num + 1)
        if render:
            self.render()

    def render(self):
        self._compute_max_label_len()
        try:
            self._ensure_next_frame()
        except _FrameAlreadyRendered:
            return

        task_ratio = self._compute_task_progress()
        field_sep = self._format_field_sep()
        icon = self._format_icon()
        task_state = [*self._format_task_state()]

        result = pt.render(
            pt.Composite(
                field_sep,
                icon,
                field_sep,
                *task_state,
                field_sep,
                *self._format_ratio_bar(task_ratio),
                field_sep,
                *self._format_labels(),
            ),
            renderer=self._renderer,
        )
        left_part_len = len(self.FIELD_SEP) + len(icon) + sum(map(len, task_state))

        if self._should_persist():
            delta_str = pt.format_time_ns(time.monotonic_ns() - self._created_at)
            result_nofmt = pt.render(
                pt.Composite(
                    pt.fit(f"+{delta_str}", left_part_len + 1, ">"),
                    self._format_field_sep(raw=True),
                    *self._format_ratio_bar(task_ratio, raw=True),
                    self._format_field_sep(raw=True),
                    *self._format_labels(raw=True),
                ),
                renderer=self._renderer,
            )
            self._echo(result_nofmt, persist=True)
        self._echo(result)

    def _echo(self, result: str, persist=False):
        if self.is_format_allowed:
            # firstly set cursor X to 0, then clear that line,
            # then echo the result, then clear that line again
            result = self.OUT_START + result + self.CSI_EL0 + self.SGR_RESET

        if self._output_buffer.getvalue():
            # something already waiting in buffer, no need to persist progress:
            self._output_buffer.write(result)
            self._io.write(self._output_buffer.popvalue().rstrip())
            self._update_last_persist_ts()
        else:
            self._io.write(result)

        if persist:
            self._io.write("\n")
            self._update_last_persist_ts()

        self._io.flush()

    def close(self):
        self._task_num = self._get_max_task_num()
        self._steps_amount = 0

        if self._output_buffer:
            self._echo("", persist=True)
            self._output_buffer.close()
            self._output_buffer = None

    def _get_max_step_num(self) -> int:
        return self._steps_amount

    def _get_max_task_num(self) -> int:
        return self._tasks_amount

    def _get_max_task_num_len(self) -> int:
        return len(str(self._get_max_task_num()))

    def _compute_task_progress(self) -> float:
        if not self._get_max_step_num():
            return 0.0
        return (self._step_num - 1) / self._get_max_step_num()

    def _compute_max_label_len(self):
        if not self._should_query_term_width():
            return

        field_seps_len = 4 * len(self.FIELD_SEP)
        icon_len = len(self.ICON)
        task_state_len = 2 * self._get_max_task_num_len() + len(self.NUM_DELIM)
        task_bar_len = self.BAR_WIDTH + len(self.BORDER_LEFT_CHAR + self.BORDER_RIGHT_CHAR)

        self._max_label_len = pt.get_terminal_width() - (
            field_seps_len + icon_len + task_bar_len + task_state_len + self.LABEL_PAD
        )

    def _ensure_next_frame(self):
        now = time.monotonic_ns()
        if not self._next_frame_ts:
            self._next_frame_ts = now

        if now >= self._next_frame_ts:
            self._next_frame_ts = now + self.FRAME_INTERVAL_SEC
            self._icon_frame += 1
        else:
            raise _FrameAlreadyRendered

    def _should_persist(self) -> bool:
        if self._last_persist_ts is None:
            return True  # first render
        from_last_persist = (time.monotonic_ns() - self._last_persist_ts) / 1e9
        return from_last_persist >= self.PERSIST_MIN_INTERVAL_SEC

    def _update_last_persist_ts(self):
        self._last_persist_ts = time.monotonic_ns()

    def _should_query_term_width(self) -> bool:
        now = time.monotonic_ns()
        if self._last_term_width_query_ts:
            if now - self._last_term_width_query_ts < self.TERM_WIDTH_QUERY_INTERVAL_SEC:
                return False

        self._last_term_width_query_ts = now
        return True

    def _format_field_sep(self, raw=False) -> str:
        if not self.is_format_allowed or raw:
            return self.FIELD_SEP
        return f"{self._styles.DEFAULT.bg.to_sgr(pt.ColorTarget.BG)}{self.FIELD_SEP}"

    def _format_ratio_bar(self, ratio: float, raw=False) -> typing.Iterable[str | pt.Fragment]:
        filled_length = math.floor(ratio * self.BAR_WIDTH)
        ratio_label = list(f"{100*ratio:>3.0f}%")
        ratio_label_len = 4  # "100%"
        ratio_label_left_pos = (self.BAR_WIDTH - ratio_label_len) // 2
        ratio_label_perc_pos = ratio_label_left_pos + 3

        if not self.is_format_allowed or raw:
            bar_chars = filled_length * FULL_BLOCK
            bar_chars += get_partial_hblock(ratio - filled_length * self.BAR_WIDTH)
            yield self.BORDER_LEFT_CHAR
            yield pt.fit(bar_chars, self.BAR_WIDTH)
            yield from ratio_label
            return

        bar_styles = [
            self._renderer.render("\x00", self._styles.BAR_FILLED).split("\x00", 1)[0],
            pt.SeqIndex.INVERSED.assemble(),
        ]
        label_styles = [
            pt.SeqIndex.BOLD.assemble(),
            pt.SeqIndex.DIM.assemble(),
        ]

        cursor = 0
        yield pt.Fragment(self.BORDER_LEFT_CHAR, self._styles.BAR_BORDER)
        yield bar_styles.pop(0)

        while cursor < self.BAR_WIDTH:
            if cursor >= filled_length and bar_styles:
                yield bar_styles.pop()
            if cursor >= ratio_label_left_pos:
                if len(label_styles) == 2:
                    yield label_styles.pop(0)
                if cursor >= ratio_label_perc_pos and label_styles:
                    yield label_styles.pop()
                if len(ratio_label):
                    cursor += 1
                    yield ratio_label.pop(0)
                    continue
            cursor += 1
            yield " "

        if bar_styles:
            yield bar_styles.pop()
        yield pt.SeqIndex.INVERSED_OFF.assemble()
        yield pt.Fragment(self.BORDER_RIGHT_CHAR, self._styles.BAR_BORDER)
        yield pt.SeqIndex.BOLD_DIM_OFF.assemble()

    def _format_icon(self, raw=False) -> pt.Fragment | str:
        if not self.is_format_allowed or raw:
            return " "
        icon = (self.ICON, " ")[self._icon_frame % 2]
        return pt.Fragment(icon, self._styles.ICON)

    def _format_task_state(self) -> typing.Iterable[pt.Fragment]:
        task_num_cur = f"{self._task_num:>{self._get_max_task_num_len()}d}"
        task_num_max = f"{self._get_max_task_num():<d}"

        yield pt.Fragment(task_num_cur, self._styles.TASK_NUM_CUR)
        yield pt.Fragment(self.NUM_DELIM, self._styles.TASK_DELIM)
        yield pt.Fragment(task_num_max, self._styles.TASK_NUM_MAX)

    def _format_labels(self, raw=False) -> typing.Iterable[pt.Fragment]:
        task_label = self._task_label
        step_label = self._step_label
        step_num = ""
        if self._print_step_num:
            step_num = f"[{self._step_num}/{self._steps_amount}] "

        # expand right label to max minus (initial) left
        label_right_text = pt.fit(
            step_label,
            self._max_label_len - self.LABEL_PAD * 2 - len(task_label) - len(step_num),
            "<",
        )
        if not self.is_format_allowed or raw:
            yield from [task_label, pt.pad(self.LABEL_PAD), step_num, label_right_text]
            return
        yield pt.Fragment(f"{task_label}{pt.pad(self.LABEL_PAD)}")
        yield pt.Fragment(f"{pt.SeqIndex.DIM}{step_num}")
        yield pt.Fragment(f"{label_right_text}")


class _PBarStyles(pt.Styles):
    def __init__(self, theme_color: pt.Color):
        self.THEME_COLOR = theme_color
        self.DEFAULT = pt.FrozenStyle(bg=pt.cv.GRAY_0)

        self.ICON = pt.FrozenStyle(self.DEFAULT, fg=self.THEME_COLOR)
        self.TASK_NUM_CUR = pt.FrozenStyle(self.ICON, bold=True)
        self.TASK_NUM_MAX = self.DEFAULT
        self.TASK_DELIM = pt.FrozenStyle(self.DEFAULT, dim=True)
        self.TASK_LABEL = self.DEFAULT

        self.RATIO_DIGITS = pt.FrozenStyle(bold=True)
        self.RATIO_PERC_SIGN = pt.FrozenStyle(dim=True)
        self.BAR_BORDER = pt.FrozenStyle(self.DEFAULT, fg=pt.cv.GRAY_19)
        self.BAR_FILLED = pt.FrozenStyle(fg=pt.cv.GRAY_0, bg=self.THEME_COLOR)
        self.BAR_EMPTY = pt.FrozenStyle(fg=self.THEME_COLOR, bg=pt.cv.GRAY_0)


class _FrameAlreadyRendered(Exception):
    pass


class _OutputBuffer(StringIO):
    def reset(self) -> None:
        self.truncate(0)
        self.seek(0)

    def popvalue(self) -> str:
        val = self.getvalue()
        self.reset()
        return val


# thresholds: 6
# ¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯
# pre1      post1 pre2      post2 pre3        post3 pre4      post4 pre5       post5 pre6         post6
# |>-----(1)-----||>-----(2)-----||>-----(3)-------||>----(4)------||>-----(5)------||>-----(6)-------|
# |______________|_______________|_________________|_______________|________________|_________________|
# ╹0 ''╵''''╹10 '╵''''╹20 '╵''''╹30 '╵''''╹40 '╵''''╹50 '╵''''╹60 '╵''''╹70 '╵''''╹80 '╵''''╹90 '╵''''╹100
#
#                  LABEl      IDX     RATIO
#        pre-1    prepare     0/6| == | 0%           0
#      start-1    task 1      1/6| != | 0%           1
# post-1 pre-2    task 1      1/6| == |16%           1
# post-2 pre-3    task 2      2/6      33%           2
# post-3 pre-4    task 3      3/6      50%           3
# post-4 pre-5    task 4      4/6      66%           4
# post-5 pre-6    task 5      5/6      83%           5
# post-6          task 6      6/6     100%           6
#
# ------------------------------------------------------------------------------------------------------
# FEATURE: tasks with non-linear step thresholds   @TODO @MAYBE
#
#     def _get_ratio_at(self, idx: int):
#         idx = max(0, min(len(self._thresholds) - 1, idx))
#         return self._thresholds[idx]
#
#     def _get_ratio(self):
#         left = self._get_ratio_global()
#         right = self._get_next_ratio_global()
#         return left + self._ratio_local * (right - left)



# ------------------------------------------------------------------------------
#  es7s/commons
#  (c) 2023 A. Shavykin <0.delameter@gmail.com>
# ------------------------------------------------------------------------------

from logging import getLogger
import re
import typing
from abc import ABCMeta, abstractmethod
from collections import deque
from collections.abc import Iterable
from dataclasses import dataclass
from typing import TypeVar

from pytermor import RGB


_T = TypeVar("_T")


@dataclass(frozen=True)
class GradientPoint:
    pos: float
    col: RGB

    def __post_init__(self):
        getLogger(__package__).debug(f"Created {self!r}")

    def __repr__(self):
        return f"{self.__class__.__name__}[pos={100*self.pos:8.4f}%, {self.col}]"


class GradientSegment:
    def __init__(self, positions: list[float], col_left: RGB, col_right: RGB):
        pos_left, pos_mid, pos_right = positions
        self.p_left: GradientPoint = GradientPoint(pos_left, col_left)
        self.p_right: GradientPoint = GradientPoint(pos_right, col_right)

        col_mid = self._interpolate_2p(self.p_left, self.p_right, pos_mid)
        self.p_mid: GradientPoint = GradientPoint(pos_mid, col_mid)
        getLogger(__package__).debug(f"Created {self!r}")

    def interpolate(self, pos: float) -> RGB:
        if pos <= self.p_mid.pos:
            pp = self.p_left, self.p_mid
        else:
            pp = self.p_mid, self.p_right
        return self._interpolate_2p(*pp, pos=pos)

    def _interpolate_2p(self, p_left: GradientPoint, p_right: GradientPoint, pos: float) -> RGB:
        pos_rel = self._pos_to_relative(p_left.pos, p_right.pos, pos)
        return RGB.from_channels(*[(pos_rel * (cr - cl) + cl) for cl, cr in zip(p_left.col, p_right.col)])

    def _pos_to_relative(self, pos1: float, pos2: float, pos_target: float) -> float:
        return (pos_target - pos1) / (pos2 - pos1)

    def __repr__(self):
        return f"{self.__class__.__name__}[{', '.join(map(repr,[self.p_left, self.p_mid, self.p_right]))}]"


class Gradient:
    def __init__(self, segments: Iterable[GradientSegment] = None, name: str = None):
        self._segments = sorted(segments, key=lambda seg: seg.p_left.pos)
        self._name = name

    def interpolate(self, pos: float) -> RGB:
        if not self._segments:
            return RGB(0)
        idx = 0
        seg = self._segments[0]
        while idx < len(self._segments):
            seg = self._segments[idx]
            if seg.p_left.pos <= pos <= seg.p_right.pos:
                break
            idx += 1
        return seg.interpolate(pos)


class IGradientReader(metaclass=ABCMeta):
    @abstractmethod
    def make(self, data: str) -> Gradient:
        ...


class deque_ext(typing.Generic[_T], deque):
    def mpop(self, amount: int = 1) -> Iterable[_T]:
        while len(self) and amount:
            amount -= 1
            yield self.pop()

    def mpopleft(self, amount: int = 1) -> Iterable[_T]:
        while len(self) and amount:
            amount -= 1
            yield self.popleft()


class GimpGradientReader(IGradientReader):
    def make(self, data_lines: list[str]) -> Gradient:
        try:
            assert data_lines.pop(0).strip() == "GIMP Gradient", "Not a GIMP gradient format"
            gradient_name = data_lines.pop(0).partition("Name:")[2].strip()
            seg_count = int(data_lines.pop(0).strip())
            assert seg_count == len(data_lines), "Malformed gradient data (line mismatch)"
        except (AssertionError, IndexError) as e:
            raise RuntimeError("Failed to read gradient file") from e

        return Gradient(self._read(*data_lines), gradient_name)

    def _read(self, *data_lines: str) -> Iterable[GradientSegment]:
        for line in data_lines:
            if not line:
                continue
            seg_raw = re.split(r"\s+", line.strip())
            if len(seg_raw) < 11:
                continue

            seg_f = deque_ext(map(float, seg_raw))

            seg_pos = [*seg_f.mpopleft(3)]
            seg_col_left = RGB.from_ratios(*map(float, seg_f.mpopleft(3)))
            seg_f.popleft()  # alpha channel value (left)
            seg_col_right = RGB.from_ratios(*map(float, seg_f.mpopleft(3)))
            seg_f.popleft()  # alpha channel value (right)
            seg = GradientSegment(
                [*seg_pos],
                seg_col_left,
                seg_col_right,
            )
            yield seg



# ------------------------------------------------------------------------------
#  es7s/commons
#  (c) 2023 A. Shavykin <0.delameter@gmail.com>
# ------------------------------------------------------------------------------
from __future__ import annotations

import typing as t
from collections import deque
from itertools import chain
from sys import getsizeof, stderr

try:
    from reprlib import repr
except ImportError:
    pass


def total_size(o: t.Any, handlers: t.Dict[t.Any, t.Iterator] = None, verbose: bool = False) -> int:
    """
    Return the approximate memory footprint of an object and all of its contents.

    Automatically finds the contents of the following builtin containers and
    their subclasses: *tuple, list, deque, dict, set* and *frozenset*.
    To search other containers, add handlers to iterate over their contents::

        handlers = {ContainerClass: iter, ContainerClass2: ContainerClass2.get_elements}

    :param o:
    :param handlers:
    :param verbose:
    """
    # origin: https://code.activestate.com/recipes/577504/

    dict_handler = lambda d: chain.from_iterable(d.items())
    all_handlers = {
        tuple: iter,
        list: iter,
        deque: iter,
        dict: dict_handler,
        set: iter,
        frozenset: iter,
    }
    all_handlers.update(handlers or {})  # user handlers take precedence
    seen = set()  # track which object id's have already been seen
    default_size = getsizeof(0)  # estimate sizeof object without __sizeof__

    def sizeof(o):
        if id(o) in seen:  # do not double count the same object
            return 0
        seen.add(id(o))
        s = getsizeof(o, default_size)

        if verbose:
            print(s, type(o), repr(o), file=stderr)

        for typ, handler in all_handlers.items():
            if isinstance(o, typ):
                s += sum(map(sizeof, handler(o)))
                break
        return s

    return sizeof(o)



# ------------------------------------------------------------------------------
#  es7s/commons
#  (c) 2023 A. Shavykin <0.delameter@gmail.com>
# ------------------------------------------------------------------------------
from __future__ import annotations

import operator
import re
import time
from functools import reduce

import math
import typing as t
from typing import final


def Regex(
    pattern: str,
    ignorecase=False,
    verbose=False,
    dotall=False,
    multiline=False,
) -> re.Pattern[str]:
    flags = t.cast(
        int,
        reduce(
            operator.xor,
            {
                re.IGNORECASE * ignorecase,
                re.VERBOSE * verbose,
                re.DOTALL * dotall,
                re.MULTILINE * multiline,
            },
        ),
    )
    return re.compile(pattern, flags)


@final
class FinalSingleton:
    _instance: FinalSingleton

    def __init__(self):
        if hasattr(self.__class__, "_instance"):
            raise RuntimeError(f"{self.__class__.__name__} is a singleton")
        self.__class__._instance = self

    @classmethod
    def get_instance(cls: FinalSingleton, require: bool = True) -> FinalSingleton | None:
        if not hasattr(cls, "_instance"):
            if require:
                raise RuntimeError(f"{cls.__name__} is uninitialized")
            return None
        return cls._instance


def autogen(__origin):
    def wrapper(*args, **kwargs):
        if not hasattr(__origin, "__generator"):
            __origin.__generator = __origin(*args, **kwargs)
        return next(__origin.__generator)

    return wrapper


def percentile(
    N: t.Sequence[float],
    percent: float,
    key: t.Callable[[float], float] = lambda x: x,
) -> float:
    """
    Find the percentile of a list of values.

    :param N:        List of values. MUST BE already sorted.
    :param percent:  Float value from 0.0 to 1.0.
    :param key:      Optional key function to compute value from each element of N.
    """
    # origin: https://code.activestate.com/recipes/511478/

    if not N:
        raise ValueError("N should be a non-empty sequence of floats")
    k = (len(N) - 1) * percent
    f = math.floor(k)
    c = math.ceil(k)
    if f == c:
        return key(N[int(k)])
    d0 = key(N[int(f)]) * (c - k)
    d1 = key(N[int(c)]) * (k - f)
    return d0 + d1


def median(
    N: t.Sequence[float],
    key: t.Callable[[float], float] = lambda x: x,
) -> float:
    """
    Find the median of a list of values.
    Wrapper around `percentile()` with fixed ``percent`` argument (=0.5).

    :param N:    List of values. MUST BE already sorted.
    :param key:  Optional key function to compute value from each element of N.
    """
    return percentile(N, percent=0.5, key=key)


def bcs(a, b):
    while a != 0 and b != 0:
        if a > b:
            a %= b
        else:
            b %= a
    return a + b


def lcm(a, b):
    m = a * b
    while a != 0 and b != 0:
        if a > b:
            a %= b
        else:
            b %= a
    return m // (a + b)


def now() -> int:
    return int(time.time())


def nowf() -> float:
    return time.time_ns() / 1e9



# ------------------------------------------------------------------------------
#  es7s/commons
#  (c) 2022-2023 A. Shavykin <0.delameter@gmail.com>
# ------------------------------------------------------------------------------
# python classifies first three as line separators

FILE_SEPARATOR = "\x1c"
GROUP_SEPARATOR = "\x1d"
RECORD_SEPARATOR = "\x1e"
UNIT_SEPARATOR = "\x1f"

SEPARATORS = [
    FILE_SEPARATOR,
    GROUP_SEPARATOR,
    RECORD_SEPARATOR,
    UNIT_SEPARATOR,
]


def get_separator(level: int = 0) -> str:
    if level < len(SEPARATORS):
        return SEPARATORS[level]
    raise LookupError



# ------------------------------------------------------------------------------
#  es7s/commons
#  (c) 2023 A. Shavykin <0.delameter@gmail.com>
# ------------------------------------------------------------------------------
from __future__ import annotations
import typing as t


class DoublyLinkedNode:
    def __init__(self):
        self._prev: t.Self | None = None
        self._next: t.Self | None = None

    def connect(self, next: t.Self):
        self._next = next
        next._prev = self

    @property
    def prev(self) -> t.Self | None:
        return self._prev

    @property
    def next(self) -> t.Self | None:
        return self._next


_DNT = t.TypeVar("_DNT", bound=DoublyLinkedNode)


class RingList(t.Generic[_DNT]):
    def __init__(self):
        self._any: _DNT | None = None  # rings do not have a head or tail

    def insert(self, target: _DNT):
        if not self._any:
            target.connect(target)  # single ring element is linked
            self._any = target  # with itself on both sides
        else:
            target.connect(self._any.next)  # cut two connections and insert
            self._any.connect(target)  # new element into created gap
            self._any = target

    def remove(self, target: _DNT):
        for el in self:
            if el == target:
                el.prev.connect(el.next)
                if self._any == el:
                    if self._any == self._any.next:  # delete link to entrypoint
                        self._any = None  # if its getting removed
                    else:
                        self._any = el.next  # if not, pick another element as an entrypoint
                return
        raise ValueError("No such element")

    def __len__(self) -> int:
        return len([_ for _ in self])

    def __iter__(self) -> t.Iterator[_DNT]:
        if (ptr := self._any) is None:
            return
        while True:  # iterate from the entrypoint
            ptr = ptr.next  # till we encounter it again
            yield ptr
            if ptr == self._any:  # funny things will happen if the ring gets
                break  # modified in the middle of iteration process

    def __repr__(self) -> str:
        if self._any is None:
            return "[]"
        return " => ".join([f"[{rel}]" for rel in ["FIRST", *self, "LAST"]])



# ------------------------------------------------------------------------------
#  es7s/commons
#  (c) 2023 A. Shavykin <0.delameter@gmail.com>
# ------------------------------------------------------------------------------
from __future__ import annotations

import pytermor as pt


class NamedGroupsRefilter(pt.AbstractNamedGroupsRefilter):
    def _get_renderer(self) -> pt.IRenderer:
        from es7s.shared import get_stdout

        return get_stdout().renderer

    def _render(self, v: pt.RT, st: pt.FT) -> str:
        return self._get_renderer().render(v, st)


class RegexValRefilter(NamedGroupsRefilter):
    def __init__(self, pattern: pt.filter.PTT[str], val_st: pt.FT):
        super().__init__(pattern, {"val": val_st})


class Transmap(dict[int, str]):
    def __init__(self, inp: str, out: str):
        if (li := len(inp)) != (lo := len(out)):
            raise ValueError(f"Strings should have equal length ({li} != {lo})")
        self._inp_set = set(inp)

        super().__init__(str.maketrans({k: v for (k, v) in zip(inp, out)}))

    def translate(self, s: str, *, strict: bool = False) -> str:
        if strict and (miss := set(s) - self._inp_set):
            raise ValueError(
                f"String contains characters without mapping: "
                + f"{' '.join([*miss][:5])}"
                + (f" (+{len(miss)} more)" if len(miss) > 5 else "")
            )
        return s.translate(self)


SUBSCRIPT_TRANS = Transmap(
    # missing: "bcdfgqwyz" and all capitals
    "0123456789+-=()aehijklmnoprstuvx",
    "₀₁₂₃₄₅₆₇₈₉₊₋₌₍₎ₐₑₕᵢⱼₖₗₘₙₒₚᵣₛₜᵤᵥₓ",
)


SUPERSCRIPT_TRANS = Transmap(
    # missing: "SXYZ"
    "0123456789+-=()abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRTUVW",
    "⁰¹²³⁴⁵⁶⁷⁸⁹⁺⁻⁼⁽⁾ᵃᵇᶜᵈᵉᶠᵍʰⁱʲᵏˡᵐⁿᵒᵖ𐞥ʳˢᵗᵘᵛʷˣʸᶻᴬᴮꟲᴰᴱfᴳᴴᴵᴶᴷᴸᴹᴺᴼᴾꟴᴿᵀᵁⱽᵂ",
)


def to_subscript(s: str, *, strict: bool = False) -> str:
    return SUBSCRIPT_TRANS.translate(s.lower(), strict=strict)


def to_superscript(s: str, *, strict: bool = False) -> str:
    return SUPERSCRIPT_TRANS.translate(s, strict=strict)


def re_unescape(s: str) -> str:
    return s.replace("\\", "")


#
# def wrap_sgr(
#     rendered: str | list[str],
#     width: int,
#     indent_first: int = 0,
#     indent_subseq: int = 0,
# ) -> str:
#     # modified to handle ⏺ lists offsets
#
#     sgrs: list[str] = []
#
#     def push(m: t.Match):
#         sgrs.append(m.group())
#         return pt.text._PRIVATE_REPLACER
#
#     if isinstance(rendered, str):  # input can be just one paragraph
#         rendered = [rendered]
#
#     inp = "\n\n".join(rendered).split("\n\n")
#     result = ""
#     for raw_line in inp:
#         # had an inspiration and wrote it; no idea how does it work exactly, it just does
#         replaced_line = re.sub(r"(\s?\S?)((\x1b\[([0-9;]*)m)+)", push, raw_line)
#         if replaced_line.lstrip().startswith("⏺"):
#             cur_indent_first = indent_first
#             cur_indent_subseq = indent_subseq + 2
#         else:
#             cur_indent_first = indent_first
#             cur_indent_subseq = indent_subseq
#         wrapped_line = f"\n".join(
#             textwrap.wrap(
#                 replaced_line,
#                 width=width,
#                 initial_indent=(cur_indent_first * " "),
#                 subsequent_indent=(cur_indent_subseq * " "),
#             )
#         )
#         final_line = re.sub(pt.text._PRIVATE_REPLACER, lambda _: sgrs.pop(0), wrapped_line)
#         result += final_line + "\n"
#     return result
